var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _AbstractGossipLog_apply;
import { TypedEventEmitter } from "@libp2p/interface";
import { logger } from "@libp2p/logger";
import { equals, toString } from "uint8arrays";
import { hashEntry } from "@canvas-js/okra";
import { ed25519 } from "@canvas-js/signatures";
import { assert, zip, prepare, prepareMessage } from "@canvas-js/utils";
import { AbortError, MessageNotFoundError, MissingParentError } from "@canvas-js/gossiplog/errors";
import * as sync from "@canvas-js/gossiplog/sync";
import target from "#target";
import { AncestorIndex } from "./AncestorIndex.js";
import { BranchMergeIndex } from "./BranchMergeIndex.js";
import { SignedMessage } from "./SignedMessage.js";
import { decodeId, encodeId, messageIdPattern } from "./ids.js";
import { getNextClock } from "./schema.js";
import { gossiplogTopicPattern } from "./utils.js";
export class AbstractGossipLog extends TypedEventEmitter {
    constructor(init) {
        super();
        Object.defineProperty(this, "topic", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "signer", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "controller", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: new AbortController()
        });
        Object.defineProperty(this, "log", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "validatePayload", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        Object.defineProperty(this, "verifySignature", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        _AbstractGossipLog_apply.set(this, void 0);
        assert(gossiplogTopicPattern.test(init.topic), "invalid topic (must be of the form 'topic' or 'topic#hash', where topic matches [a-zA-Z0-9\\.\\-])");
        this.topic = init.topic;
        this.signer = init.signer ?? ed25519.create();
        __classPrivateFieldSet(this, _AbstractGossipLog_apply, init.apply, "f");
        this.validatePayload = init.validatePayload ?? ((payload) => true);
        this.verifySignature = init.verifySignature ?? this.signer.scheme.verify;
        this.log = logger(`canvas:gossiplog:[${this.topic}]`);
    }
    async close() {
        this.log("closing");
        this.controller.abort();
        await this.tree.close();
        await this.db.close();
    }
    async replay() {
        await this.tree.read(async (txn) => {
            for (const leaf of txn.keys()) {
                const id = decodeId(leaf);
                const record = await this.db.get("$messages", id);
                if (record === null) {
                    this.log.error("failed to get message %s from database", id);
                    continue;
                }
                const { signature, message, branch } = record;
                const signedMessage = this.encode(signature, message, { branch });
                assert(signedMessage.id === id);
                await __classPrivateFieldGet(this, _AbstractGossipLog_apply, "f").apply(this, [signedMessage]);
            }
        });
    }
    async connect(url, options = {}) {
        await target.connect(this, url, options);
    }
    async listen(port, options = {}) {
        await target.listen(this, port, options);
    }
    async startLibp2p(config) {
        const libp2p = await target.startLibp2p(this, config);
        this.controller.signal.addEventListener("abort", () => libp2p.stop());
        return libp2p;
    }
    encode(signature, message, context = {}) {
        assert(this.topic === message.topic, "expected this.topic === topic");
        const preparedMessage = prepareMessage(message);
        assert(this.validatePayload(preparedMessage.payload), "error encoding message (invalid payload)");
        return SignedMessage.encode(signature, preparedMessage, context);
    }
    decode(value, context = {}) {
        const signedMessage = SignedMessage.decode(value, context);
        assert(this.topic === signedMessage.message.topic, "expected this.topic === topic");
        assert(this.validatePayload(signedMessage.message.payload), "error decoding message (invalid payload)");
        return signedMessage;
    }
    async getClock() {
        const heads = await this.db.query("$heads", { orderBy: { id: "asc" } });
        const ids = heads.map(({ id }) => id);
        const clock = getNextClock(ids.map(encodeId));
        return [clock, ids];
    }
    async has(id) {
        assert(messageIdPattern.test(id), "invalid message ID");
        const records = await this.db.query("$messages", { select: { id: true }, where: { id } });
        return records.length > 0;
    }
    async get(id) {
        const record = await this.db.get("$messages", id);
        if (record === null) {
            return null;
        }
        const { signature, message, branch } = record;
        return this.encode(signature, message, { branch });
    }
    getMessages(range = {}) {
        const { reverse = false, limit, ...where } = range;
        return this.db.query("$messages", {
            where: { id: where },
            select: { id: true, signature: true, message: true, branch: true },
            orderBy: { id: reverse ? "desc" : "asc" },
            limit,
        });
    }
    async *iterate(range = {}) {
        const { reverse = false, limit, ...where } = range;
        for await (const row of this.db.iterate("$messages", {
            where: { id: where },
            select: { id: true, signature: true, message: true },
            orderBy: { id: reverse ? "desc" : "asc" },
            limit,
        })) {
            yield this.encode(row.signature, row.message);
        }
    }
    /**
     * Sign and append a new *unsigned* message to the end of the log.
     * The current concurrent heads of the local log are used as parents.
     */
    async append(payload, { signer = this.signer } = {}) {
        let root = null;
        let heads = null;
        const signedMessage = await this.tree.write(async (txn) => {
            const [clock, parents] = await this.getClock();
            const message = {
                topic: this.topic,
                clock,
                parents,
                payload: prepare(payload, { replaceUndefined: true }),
            };
            const signature = await signer.sign(message);
            const signedMessage = this.encode(signature, message);
            this.log("appending message %s at clock %d with parents %o", signedMessage.id, clock, parents);
            const result = await this.apply(txn, signedMessage);
            root = result.root;
            heads = result.heads;
            return signedMessage;
        });
        assert(root !== null && heads !== null, "failed to commit transaction");
        this.dispatchEvent(new CustomEvent("commit", { detail: { root, heads } }));
        return signedMessage;
    }
    /**
     * Insert an existing signed message into the log (ie received via HTTP API).
     * If any of the parents are not present, throw an error.
     */
    async insert(signedMessage) {
        const { message, signature } = signedMessage;
        assert(message.topic === this.topic, `expected message.topic === this.topic`);
        await this.verifySignature(signature, message);
        const { clock, parents } = signedMessage.message;
        const signedMessageInstance = signedMessage instanceof SignedMessage ? signedMessage : this.encode(signature, message);
        const id = signedMessageInstance.id;
        this.log("inserting message %s at clock %d with parents %o", id, clock, parents);
        const parentKeys = message.parents.map(encodeId);
        const result = await this.tree.write(async (txn) => {
            if (txn.has(signedMessage.key)) {
                return null;
            }
            for (const parentKey of parentKeys) {
                const leaf = txn.getNode(0, parentKey);
                if (leaf === null) {
                    const parent = decodeId(parentKey);
                    this.log.error("missing parent %s of message %s: %O", parent, id, message);
                    throw new MissingParentError(parent, id);
                }
            }
            return await this.apply(txn, signedMessageInstance);
        });
        if (result !== null) {
            this.dispatchEvent(new CustomEvent("commit", { detail: result }));
        }
        return { id };
    }
    async apply(txn, signedMessage) {
        const { id, signature, message, key, value } = signedMessage;
        this.log.trace("applying %s %O", id, message);
        const parentMessageRecords = [];
        for (const parentId of message.parents) {
            const parentMessageRecord = await this.db.get("$messages", parentId);
            if (parentMessageRecord === null) {
                throw new Error(`missing parent ${parentId} of message ${id}`);
            }
            parentMessageRecords.push(parentMessageRecord);
        }
        const branch = await this.getBranch(id, parentMessageRecords);
        signedMessage.branch = branch;
        await __classPrivateFieldGet(this, _AbstractGossipLog_apply, "f").apply(this, [signedMessage]);
        const hash = toString(hashEntry(key, value), "hex");
        const branchMergeIndex = new BranchMergeIndex(this.db);
        for (const parentMessageRecord of parentMessageRecords) {
            if (parentMessageRecord.branch !== branch) {
                await branchMergeIndex.insertBranchMerge({
                    source_branch: parentMessageRecord.branch,
                    source_clock: parentMessageRecord.clock,
                    source_message_id: parentMessageRecord.id,
                    target_branch: branch,
                    target_clock: message.clock,
                    target_message_id: id,
                });
            }
        }
        const messageRecord = { id, signature, message, hash, branch, clock: message.clock };
        const heads = await this.db
            .query("$heads", { select: { id: true } })
            .then((heads) => heads.map((head) => head.id))
            .then((heads) => heads.filter((head) => message.parents.includes(head)));
        await this.db.apply([
            ...heads.map((head) => ({ model: "$heads", operation: "delete", key: head })),
            { model: "$heads", operation: "set", value: { id } },
            { model: "$messages", operation: "set", value: messageRecord },
        ]);
        txn.set(key, value);
        await new AncestorIndex(this.db).indexAncestors(id, message.parents);
        this.dispatchEvent(new CustomEvent("message", { detail: signedMessage }));
        const root = txn.getRoot();
        return { root, heads };
    }
    async newBranch() {
        const maxBranchRecords = await this.db.query("$messages", {
            select: { id: true, branch: true },
            limit: 1,
            orderBy: { branch: "desc" },
        });
        if (maxBranchRecords.length === 0) {
            return 0;
        }
        else {
            return maxBranchRecords[0].branch + 1;
        }
    }
    async getBranch(messageId, parentMessages) {
        if (parentMessages.length === 0) {
            return await this.newBranch();
        }
        const parentMessageWithMaxClock = parentMessages.reduce((max, parentMessage) => max.branch > parentMessage.branch ? max : parentMessage);
        const branch = parentMessageWithMaxClock.branch;
        const messagesAtBranchClockPosition = await this.db.query("$messages", {
            select: { id: true, branch: true, clock: true },
            where: {
                branch,
                clock: { gt: parentMessageWithMaxClock.clock },
                id: { neq: messageId },
            },
            limit: 1,
        });
        if (messagesAtBranchClockPosition.length > 0) {
            return await this.newBranch();
        }
        else {
            return branch;
        }
    }
    async isAncestor(id, ancestor, visited = new Set()) {
        return await new AncestorIndex(this.db).isAncestor(id, ancestor, visited);
    }
    /**
     * Sync with a remote source, applying and inserting all missing messages into the local log
     */
    async sync(snapshot, options = {}) {
        const start = performance.now();
        let messageCount = 0;
        let complete = true;
        const root = await this.tree.read(async (txn) => {
            const driver = new sync.Driver(this.topic, snapshot, txn);
            try {
                for await (const keys of driver.sync()) {
                    const values = await snapshot.getValues(keys);
                    for (const [key, value] of zip(keys, values)) {
                        const signedMessage = this.decode(value, {
                            source: options.peer === undefined ? undefined : { type: "sync", peer: options.peer },
                        });
                        assert(equals(key, signedMessage.key), "invalid message key");
                        await this.insert(signedMessage);
                        messageCount++;
                    }
                }
            }
            catch (err) {
                if (err instanceof AbortError) {
                    complete = false;
                }
                else {
                    throw err;
                }
            }
            return txn.getRoot();
        });
        const duration = Math.ceil(performance.now() - start);
        if (complete) {
            this.log("completed sync with %s (%d messages in %dms)", options.peer ?? "unknown", messageCount, duration);
        }
        else {
            this.log("aborted sync with %s (%d messages in %dms)", options.peer ?? "unknown", messageCount, duration);
        }
        this.dispatchEvent(new CustomEvent("sync", { detail: { peer: options.peer, messageCount, duration } }));
        return { complete, messageCount };
    }
    serve(callback) {
        return this.tree.read((txn) => callback({
            getRoot: () => txn.getRoot(),
            getNode: (level, key) => txn.getNode(level, key),
            getChildren: (level, key) => txn.getChildren(level, key),
            getValues: async (keys) => {
                const values = [];
                const ids = keys.map(decodeId);
                const messageRecords = await this.db.getMany("$messages", ids);
                for (let i = 0; i < messageRecords.length; i++) {
                    const messageRecord = messageRecords[i];
                    if (messageRecord === null) {
                        throw new MessageNotFoundError(ids[i]);
                    }
                    const { signature, message, branch } = messageRecord;
                    const signedMessage = this.encode(signature, message, { branch });
                    assert(equals(signedMessage.key, keys[i]), "invalid message key");
                    values.push(signedMessage.value);
                }
                return values;
            },
        }));
    }
}
_AbstractGossipLog_apply = new WeakMap();
Object.defineProperty(AbstractGossipLog, "schema", {
    enumerable: true,
    configurable: true,
    writable: true,
    value: {
        $messages: {
            id: "primary",
            signature: "json",
            message: "json",
            hash: "string",
            branch: "integer",
            clock: "integer",
            $indexes: ["branch", "clock"],
        },
        $heads: { id: "primary" },
        ...AncestorIndex.schema,
        ...BranchMergeIndex.schema,
    }
});
